"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const client_1 = require("../src/client");
const constants_1 = require("../src/constants");
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const USER_KEYPAIR = process.env.USER_KEYPAIR_OVERRIDE || process.env.MB_PAYER_KEYPAIR;
const MANGO_ACCOUNT_PK = process.env.MANGO_ACCOUNT_PK;
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
async function settlePnlForMangoAccount(client, group, mangoAccount, settler) {
    for (const pp of mangoAccount.perpActive()) {
        // settle only a specific market
        // if (pp.marketIndex != 2) continue;
        const pm = group.getPerpMarketByMarketIndex(pp.marketIndex);
        const upnlUi = pp.getUnsettledPnlUi(pm);
        if (upnlUi > 0) {
            const c = await pm.getSettlePnlCandidates(client, group, undefined, 'negative');
            const sig = await client.perpSettlePnl(group, mangoAccount, c[0].account, settler, pm.perpMarketIndex);
            console.log(`Settled ${mangoAccount.publicKey}'s pnl for ${pm.name} at ${sig}`);
        }
        else {
            const c = await pm.getSettlePnlCandidates(client, group, undefined, 'positive');
            const sig = await client.perpSettlePnl(group, c[0].account, mangoAccount, settler, pm.perpMarketIndex);
            console.log(`Settled ${mangoAccount.publicKey}'s pnl for ${pm.name} at ${sig}`);
        }
    }
}
async function main() {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const kp = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(USER_KEYPAIR, 'utf-8'))));
    const wallet = new anchor_1.Wallet(kp);
    const provider = new anchor_1.AnchorProvider(connection, wallet, options);
    const client = client_1.MangoClient.connect(provider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'api',
    });
    const group = await client.getGroup(new web3_js_1.PublicKey('78b8f4cGCwmZ9ysPFMWLaLTkkaYnUjwMJYStWe5RTSSX'));
    let mangoAccounts = await client.getAllMangoAccounts(group, true);
    const settler = await client.getMangoAccountForOwner(group, wallet.publicKey, 0, true);
    if (MANGO_ACCOUNT_PK) {
        const mangoAccount = await client.getMangoAccount(new web3_js_1.PublicKey(MANGO_ACCOUNT_PK));
        await settlePnlForMangoAccount(client, group, mangoAccount, settler);
        process.exit();
    }
    mangoAccounts = mangoAccounts.filter((a) => a.perpActive().length > 0 &&
        a
            .perpActive()
            .some((pp) => pp.getBasePositionUi(group.getPerpMarketByMarketIndex(pp.marketIndex)) == 0 &&
            Math.abs(pp.getUnsettledFundingUi(group.getPerpMarketByMarketIndex(pp.marketIndex))) > 0) === true);
    for (const mangoAccount of mangoAccounts) {
        await settlePnlForMangoAccount(client, group, mangoAccount, settler);
    }
    process.exit();
}
main();
